#include <unistd.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <sys/time.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/hdreg.h>
#include <errno.h>
#include <sys/ioctl.h>

#include "mct_test.h"

#define TIMING_BUF_MB		2 //incorrect because time = 0.1
#define TIMING_BUF_BYTES	(TIMING_BUF_MB * 1024 * 1024)

#define O_DIRECT	 040000	/* Direct disk access.	*/

extern int verbose;
extern char dev_file_path[1024];

int rc = 0;

int do_drive_cmd (int fd, unsigned char *args, unsigned int timeout_secs)
{
#ifdef SG_IO

	struct ata_tf tf;
	void *data = NULL;
	unsigned int data_bytes = 0;
	int rc;

	if (args == NULL)
		goto use_legacy_ioctl;
	/*
	 * Reformat and try to issue via SG_IO:
	 * args[0]: command in; status out.
	 * args[1]: lbal for SMART, nsect for all others; error out
	 * args[2]: feat in; nsect out.
	 * args[3]: data-count (512 multiple) for all cmds.
	 */
	tf_init(&tf, args[0], 0, 0);
	tf.lob.nsect = args[1];
	tf.lob.feat  = args[2];
	if (args[3]) {
		data_bytes   = args[3] * 512;
		data         = args + 4;
		if (!tf.lob.nsect)
			tf.lob.nsect = args[3];
	}
	if (tf.command == ATA_OP_SMART) {
		tf.lob.nsect = args[3];
		tf.lob.lbal  = args[1];
		tf.lob.lbam  = 0x4f;
		tf.lob.lbah  = 0xc2;
	}

	rc = sg16(fd, SG_READ, is_dma(tf.command), &tf, data, data_bytes, timeout_secs);
	if (rc == -1) {
		if (errno == EINVAL || errno == ENODEV || errno == EBADE)
			goto use_legacy_ioctl;
	}

	if (rc == 0 || errno == EIO) {
		args[0] = tf.status;
		args[1] = tf.error;
		args[2] = tf.lob.nsect;
	}
	return rc;

use_legacy_ioctl:
#endif /* SG_IO */
	return ioctl(fd, HDIO_DRIVE_CMD, args);
}

static int seek_to_zero (int fd)
{
	if (lseek(fd, (off_t) 0, SEEK_SET)) {
		myprintf("lseek() failed");
		return 1;
	}
	return 0;
}

static void flush_buffer_cache (int fd)
{
	sync();
	fsync(fd);				/* flush buffers */
	fdatasync(fd);				/* flush buffers */
	sync();
	if (ioctl(fd, BLKFLSBUF, NULL)) {	/* do it again, big time */
		myprintf("BLKFLSBUF failed");
	} else {
		do_drive_cmd(fd, NULL, 0);	/* IDE: await completion */
	}
	sync();
}

static void *prepare_timing_buf (unsigned int len)
{
	unsigned int i;
	__u8 *buf;

	buf = mmap(NULL, len, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0);
	if (buf == MAP_FAILED) {
		myprintf("could not allocate timing buf");
		return NULL;
	}
	for (i = 0; i < len; i += 4096)
		buf[i] = 0; /* guarantee memory is present/assigned */
	if (-1 == mlock(buf, len)) {
		myprintf("mlock() failed on timing buf");
		munmap(buf, len);
		return NULL;
	}
	mlockall(MCL_CURRENT|MCL_FUTURE); // don't care if this fails on low-memory machines
	sync();

	/* give time for I/O to settle */
	//sleep(1);
	return buf;
}

static int read_big_block (int fd, char *buf)
{
	int i;
	if ((rc = read(fd, buf, TIMING_BUF_BYTES)) != TIMING_BUF_BYTES) {
		if (rc) {
			if (rc == -1) {
				myprintf("read() failed");
			} else {
				myprintf("read(%u) returned %u bytes\n", TIMING_BUF_BYTES, rc);
			}
		} else {
			myprintf("read() hit EOF - device too small\n");
		}
		return 1;
	}
	/* access all sectors of buf to ensure the read fully completed */
	for (i = 0; i < TIMING_BUF_BYTES; i += 512)
		buf[i] &= 1;

	return 0;
}

static int time_cache (int fd)
{
	char *buf;
	struct itimerval e1, e2;
	double elapsed, elapsed2;
	unsigned int iterations, total_MB = 0;

	buf = prepare_timing_buf(TIMING_BUF_BYTES);
	if (!buf)
		return TEST_FAIL;

	/*
	 * getitimer() is used rather than gettimeofday() because
	 * it is much more consistent (on my machine, at least).
	 */
	setitimer(ITIMER_REAL, &(struct itimerval){{1000,0},{1000,0}}, NULL);
	if (seek_to_zero (fd)) return -1;
	if (read_big_block (fd, buf)) return -1;
	myprintf("Timing cached reads:   \n");
	fflush(stdout);

	/* Clear out the device request queues & give them time to complete */
	flush_buffer_cache(fd);
	//sleep(1);

	/* Now do the timing */
	iterations = 0;
	getitimer(ITIMER_REAL, &e1);
	do {
		++iterations;
		if (seek_to_zero (fd) || read_big_block (fd, buf))
			goto quit;
		getitimer(ITIMER_REAL, &e2);
		elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
			+ ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
	} while (elapsed < 0.1);
	total_MB = iterations * TIMING_BUF_MB;

	elapsed = (e1.it_value.tv_sec - e2.it_value.tv_sec)
		+ ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);

	/* Now remove the lseek() and getitimer() overheads from the elapsed time */
	getitimer(ITIMER_REAL, &e1);
	do {
		if (seek_to_zero (fd))
			goto quit;
		getitimer(ITIMER_REAL, &e2);
		elapsed2 = (e1.it_value.tv_sec - e2.it_value.tv_sec)
			+ ((e1.it_value.tv_usec - e2.it_value.tv_usec) / 1000000.0);
	} while (--iterations);

	elapsed -= elapsed2;

	if (total_MB >= elapsed) { // more than 1MB/s
		myprintf("%3u MB in %5.2f seconds = %6.2f MB/sec\n",
			 total_MB, elapsed,
			 total_MB / elapsed);
	} else {
		myprintf("%3u MB in %5.2f seconds = %6.2f kB/sec\n",
			 total_MB, elapsed,
			 total_MB / elapsed * 1024);
	}

	flush_buffer_cache(fd);
	//sleep(1);
quit:
	munlockall();
	munmap(buf, TIMING_BUF_BYTES);

	if (total_MB > 0)
		return 0;
	return TEST_FAIL;
}

int run_hdd_test() {

	int fd, res;

	fd = open(dev_file_path, O_RDONLY|O_NONBLOCK);
	if (fd < 0)
		return TEST_FAIL;

	myprintf("run HDD test...\n");
	res = time_cache(fd);

	return res;
}
